/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN:ENUM,INT (../aisleriot/src/lib/ar-marshal.list:1) */
G_GNUC_INTERNAL
void ar_marshal_BOOLEAN__ENUM_INT (GClosure     *closure,
                                   GValue       *return_value,
                                   guint         n_param_values,
                                   const GValue *param_values,
                                   gpointer      invocation_hint,
                                   gpointer      marshal_data);

/* BOOLEAN:STRING,UINT,ENUM (../aisleriot/src/lib/ar-marshal.list:2) */
G_GNUC_INTERNAL
void ar_marshal_BOOLEAN__STRING_UINT_ENUM (GClosure     *closure,
                                           GValue       *return_value,
                                           guint         n_param_values,
                                           const GValue *param_values,
                                           gpointer      invocation_hint,
                                           gpointer      marshal_data);

/* VOID:STRING,STRING (../aisleriot/src/lib/ar-marshal.list:3) */
G_GNUC_INTERNAL
void ar_marshal_VOID__STRING_STRING (GClosure     *closure,
                                     GValue       *return_value,
                                     guint         n_param_values,
                                     const GValue *param_values,
                                     gpointer      invocation_hint,
                                     gpointer      marshal_data);


G_END_DECLS
