/*
 * games-sound.h: common sound player for gnome-games
 *
 * Copyright © 2007-2008 Andreas Røsdal
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#ifndef GAMES_SOUND_H
#define GAMES_SOUND_H

#include <gtk/gtk.h>

G_BEGIN_DECLS

gboolean ar_sound_is_available (void);
void ar_sound_init             (GdkScreen *screen);
#ifndef GDK_MULTIHEAD_SAFE
void ar_sound_play             (const gchar *sound_name);
#endif
void ar_sound_play_for_screen  (const gchar *sound_name,
                                GdkScreen *screen);
void ar_sound_play_for_event   (const gchar *sound_name,
                                GdkEvent *event);
void ar_sound_play_for_widget  (const gchar *sound_name,
                                GtkWidget *widget);
void ar_sound_enable           (gboolean enabled);
gboolean ar_sound_is_enabled   (void);

G_END_DECLS

#endif /* !GAMES_SOUND_H */
